/*
 * Copyright (C) 2019 - 2023 Devexperts Solutions IE Limited
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
 * If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
import { DataSeriesModel } from '../../model/data-series.model';
import { Unit, unitToPercent } from '../../model/scaling/viewport.model';
import { PriceIncrementsUtils } from '../../utils/price-increments.utils';
import { YExtentComponent } from '../pane/extent/y-extent-component';
import { YExtentFormatters } from '../pane/pane.component';

export const createRegularPriceFormatter =
	(extent: YExtentComponent) =>
	(value: unknown): string => {
		let checkedValue: number;
		if (typeof value === 'number') {
			checkedValue = value;
		} else if (typeof value === 'string') {
			checkedValue = Number(value);
		} else {
			return '—';
		}

        const checkedValueAbs = Math.abs(checkedValue);
        if (checkedValueAbs>=1e9) return (Math.round(checkedValue / 1e7) / 1e2).toFixed(1) + "b";
        if (checkedValueAbs>=1e6) return (Math.round(checkedValue / 1e4) / 1e2).toFixed(1) + "m";
        if (checkedValueAbs>=1e5) return checkedValue.toFixed(1);

		const [mainDataSeries] = extent.dataSeries;
		if (mainDataSeries !== undefined) {
			//const precision = PriceIncrementsUtils.getPricePrecision(checkedValue, mainDataSeries.pricePrecisions);
            const getPricePrecision = (v: number) => {
                if (!isNaN(v)) {
                    const va = Math.abs(v);
                    if (va<0.00000001) {
                        // essentially zero
                        return 2;
                    } else if (va<0.0001) {
                        return 8;
                    } else if (va<0.01) {
                        return 6;
                    } else if (va<10) {
                        return 4;
                    } else if (va>=10 && va<100) {
                        return 3;
                    }
                }
                return 2;
            }
			return checkedValue.toFixed(getPricePrecision(checkedValue));
		}
		return `${checkedValue}`;
	};

export const createPercentFormatter =
	(extent: YExtentComponent) =>
	(value: Unit, dataSeries?: DataSeriesModel): string => {
		const [mainDataSeries] = extent.dataSeries;
		let valueUnit = value;
		const series = dataSeries ?? mainDataSeries;
		if (series !== undefined) {
			valueUnit = unitToPercent(value, series.getBaseline());
		}
		// always apply default precision for percent
		const formatted = valueUnit.toFixed(PriceIncrementsUtils.DEFAULT_PRECISION).replace('-', '−') + ' %';
		return formatted === '−0.00 %' ? '0.00 %' : formatted;
	};

export const createBlankPriceFormatter =
    (extent: YExtentComponent) =>
    (value: unknown): string => {
        return "";
    };

export const createYExtentFormatters = (extent: YExtentComponent): YExtentFormatters => ({
	percent: createPercentFormatter(extent),
	regular: createRegularPriceFormatter(extent),
    blank: createBlankPriceFormatter(extent)
});
